/* This program is placed in the public domain, you may do whatever you wish with the code.
 * 
 * Authors: Chris Savvopoulos (glouk), Andreas Liapis (wcf)
 * GNU/Linux port: John Tsiombikas (Nuclear)
 */
#include "Raytracer.h"

const Vector Lights[4] = {	{ 0,0,0},
							{ 0.577350f, 0.577350f, 0.577350f } ,
							{ 0.707106f, -0.707106f, 0 } ,
							{ 0,0,1 }
};
Sphere Spheres[7] = {
	{},
	{ {}, 360, 274, 260, 80 },
	{ 100, 0, 0, {}, 36 },
	{ 0, 100, 0, {}, 36 },
	{ 0, 0, 100, {}, 36 },
	{ 100,100,0, {}, 36 },
	{ 100,0,100, {}, 36 }



};


void Raycast (Ray *r, Vector *col, uint MaxDepth)
{
	if (!MaxDepth)
		return;
	
	int SphereID=0;
	float t=10000000.0f,tmp,tmp2,tmp3,tmp4;

	Vector N,R;
	Ray r2;
	Vector col2;
	
	uint s;
	for (s=6; s; s--)
	{
		tmp = r->a.x-Spheres[s].Position.x;
		tmp2 = r->a.y-Spheres[s].Position.y;
		tmp3 = r->a.z-Spheres[s].Position.z;

		tmp4 = tmp*tmp + tmp2*tmp2 + tmp3*tmp3 - Spheres[s].Radius*Spheres[s].Radius;
		tmp = r->v.x*tmp + r->v.y*tmp2 + r->v.z*tmp3;
		tmp2 = r->v.x*r->v.x + r->v.y*r->v.y + r->v.z*r->v.z;
		tmp3 = tmp*tmp-tmp2*tmp4;

		
		if (tmp3 >= 0)
		{
#ifdef USINGLIBS
			tmp = -(sqrtf(tmp3)+tmp)/tmp2;
#else
#ifdef __GNUC__
			asm volatile(
				"fld %1\n\t"
				"fsqrt\n\t"
				"fadd %2\n\t"
				"fchs\n\t"
				"fdiv %3\n\t"
				"fstp %0\n\t"
				: "=m"(tmp)
				: "m"(tmp3), "m"(tmp), "m"(tmp2)
			);
#else	// assume MSVC
			__asm
			{
				fld tmp3
				fsqrt
				fadd tmp
				fchs
				fdiv tmp2
				fstp tmp
			}
#endif	// __GNUC__
#endif
			if (tmp > 0.0f && tmp < t)
			{
				SphereID = s;
				t = tmp;
			}
		}
		
	}
	
	if (!SphereID)
		return;
	
	r2.a.x = r->a.x + t * r->v.x;
	r2.a.y = r->a.y + t * r->v.y;
	r2.a.z = r->a.z + t * r->v.z;

	tmp = 1 / Spheres[SphereID].Radius;
	N.x = (Spheres[SphereID].Position.x-r2.a.x)*tmp;
	N.y = (Spheres[SphereID].Position.y-r2.a.y)*tmp;
	N.z = (Spheres[SphereID].Position.z-r2.a.z)*tmp;

#ifdef USINGLIBS
	tmp = 1/sqrtf(r->v.x*r->v.x + r->v.y*r->v.y + r->v.z*r->v.z);
#else
	tmp = r->v.x*r->v.x + r->v.y*r->v.y + r->v.z*r->v.z;

#ifdef __GNUC__
	asm volatile(
		"fld %1\n\t"
		"fsqrt\n\t"
		"fstp %0\n\t"
		: "=m"(tmp)
		: "m"(tmp)
	);
#else	// MSVC
	__asm
	{
		fld tmp
		fsqrt
		fstp tmp
	}
	tmp=1/tmp;
#endif	// __GNUC__
#endif

	r->v.x *= tmp;
	r->v.y *= tmp;
	r->v.z *= tmp;

	uint l;
	for (l=3; l; l--)
	{
		r2.v.x = -Lights[l].x;
		r2.v.y = -Lights[l].y;
		r2.v.z = -Lights[l].z;
		r2.b.x = r2.a.x+r2.v.x;
		r2.b.y = r2.a.y+r2.v.y;
		r2.b.z = r2.a.z+r2.v.z;
		
		tmp = N.x*Lights[l].x + N.y*Lights[l].y + N.z*Lights[l].z;

		tmp2 = 2.0f*(N.x*Lights[l].x + N.y*Lights[l].y + N.z*Lights[l].z);
		R.x = N.x*tmp2 - Lights[l].x;
		R.y = N.y*tmp2 - Lights[l].y;
		R.z = N.z*tmp2 - Lights[l].z;
		tmp2 = r->v.x*R.x + r->v.y*R.y + r->v.z*R.z;
		
		if (tmp < 0)
			tmp = 0;
		
		if (tmp2 < 0)
			tmp2 = 0;
		else
		{
			tmp2 *= tmp2;
			tmp2 *= tmp2;
			tmp2 *= tmp2;
			tmp2 *= tmp2;
			tmp2 *= tmp2;
			tmp2 *= tmp2;
		}
		tmp = 80*(tmp + tmp2);
		col->x += tmp;
		col->y += tmp;
		col->z += tmp;
	}

	col->x += Spheres[SphereID].Color.x;
	col->y += Spheres[SphereID].Color.y;
	col->z += Spheres[SphereID].Color.z;

	r->v.x = -r->v.x; r->v.y = -r->v.y; r->v.z = -r->v.z; 
	tmp = 2.0f*(N.x*r->v.x + N.z*r->v.z + N.z*r->v.z);
	r2.v.x = N.x*tmp - r->v.x;
	r2.v.y = N.y*tmp - r->v.y;
	r2.v.z = N.z*tmp - r->v.z;

	r2.b.x = r2.a.x + r2.v.x;
	r2.b.y = r2.a.y + r2.v.y;
	r2.b.z = r2.a.z + r2.v.z;
	
	col2.x = col2.y = col2.z = 0;
	Raycast (&r2, &col2, MaxDepth-1);
	
	col->x += 0.5f*col2.x;
	col->y += 0.5f*col2.y;
	col->z += 0.5f*col2.z;
}

