/* This program is placed in the public domain, you may do whatever you wish with the code.
 * 
 * Authors: Chris Savvopoulos (glouk), Andreas Liapis (wcf)
 * GNU/Linux port: John Tsiombikas (Nuclear)
 */
#define WIDTH 640
#define HEIGHT 480
#define FONTHEIGHT 24
#define LINES 30

#define USE_SDL

#include "Raytracer.h"

#ifdef USINGLIBS
#include <stdio.h>
#include <stdlib.h>
#endif

#ifdef USE_SDL
#include <SDL.h>
#include <SDL_ttf.h>

#define GetTickCount()	SDL_GetTicks()
#define ShowCursor(x)	SDL_ShowCursor(x)
#else
#include <windows.h>
#include <ddraw.h>
#include <dsound.h>
#endif	// USE_SDL

#if !defined(USINGLIBS) && defined(__linux__)
void exit(int status) {
	asm volatile(
		"mov $1, %%eax\n\t"
		"mov %0, %%ebx\n\t"
		"int $0x80\n\t"
		:: "d"(status)
	);
	for(;;);	// avoid a stupid warning
}
#endif

#if !defined(USINGLIBS) && defined(WIN32)
void exit(int status) {
	ExitProcess(status);
}
#endif

Vector Amp[5] = {
	{ 200,100,200 },
	{ 280,030,60 },
	{ 60,230,80 },
	{ 200,200,200 },
	{ 160,160,160 }
};
byte Phase[5][3] = {
	{ 0,1,1 },
	{ 1,1,0 },
	{ 0,1,0 },
	{ 0,1,0 },
	{ 1,1,0 }
};
short Off[15] = {	7000, 7000, 3000, 3000, 7000,
					7000, 5000,   0, 13000, 3000,
					3000, 4729, 4729, 8936, 8936   };

#ifdef USE_SDL
SDL_Surface *Backbuffer;
int windowed = 0;
#else
IDirectDraw7 *DDraw;
IDirectDrawSurface7 *Primary, *Backbuffer;
DDSURFACEDESC2 ScreenDesc;
#endif	// USE_SDL

#ifdef USE_SDL
TTF_Font *font;
#else
HDC FontDC;       // memory DC for rendering text
LPSTR BmpBits;     // pointer to DIB bits
#endif	// USE_SDL
char Text[LINES][71] = {
	"spherez by ARG @ React 2004\n",
	" \n",
	"glouk: graphics and synch\n",
	"wcf: music and softsynth\n",
	" \n",
	"Greetings to:\n",
	"_outsider\n",
	"Da_Medron\n",
	"iM\n",
	"imak\n",
	"Jobo\n",
	"kuifware\n",
	"moT\n",
	"NaViS\n",
	"Nuclear\n",
	"orpheas\n",
	"Pantsos\n",
	"Rawnoise\n",
	"Soulrider\n",
	"thor\n",
	"xprience\n",
	"zafos\n",
	" \n",
	"Respect to:\n",
	"asd\n",
	"exceed\n",
	"farbrausch\n",
	"federation against nature\n",
	"quadra\n",
	"the lab\n"
};

extern Sphere Spheres[7];

#if !defined(USINGLIBS) && defined(_MSC_VER)
extern "C" {
	int _fltused;
};
#endif

#define NoteLength 22100
const float TwoPi = 6.283185f / 22100.0f;

char VSin1[]= {10, 0,0,-1, 12,-2 ,0 ,-1 ,10, 0,2,-1,12 ,-2,0 ,0 ,0,0 ,0,10, 0,0,-1, 12,-2 ,0 ,-1 ,10, 0,0,-1,12 ,10, 0,0,-1,12,-2 ,0 ,-1 ,10, 0,0,-1,12 ,10, 0,0,-1, 12,-2 ,0 ,-1	,10, 0,0,-1,12 ,10, 0,0,-1, 12,-2 ,0 ,-1 ,10, 0,0,-1,12 ,10, 0,0,-1, 12,-2 ,0 ,-1 ,10, 0,0,-1,12 ,10, 0,0,-1, 12,-2 ,0 ,-1 ,10, 0,0,-1,12 ,10, 0,0,-1, 12,-2 ,0 ,-1 ,10, 0,0,-1,12 ,10, 0,0,-1, 12,-2 ,0 ,-1 ,10, 0,0,-1,12 ,10, 0,0,-1, 12,-2 ,0 ,-1 ,10, 0,0,-1,12 ,10, 0,0,-1, 12,-2 ,0 ,-1 ,10, 0,0,-1,12 ,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
char VSin2[]= {12,-2,0,-1, 10, 0 ,0 ,-1 ,12,-2,2,-1,10 , 0,0 ,0 ,0,0 ,0,12,-2,0,-1, 10, 0 ,0 ,-1 ,12,-2,0,-1,10 ,12,-2,0,-1, 10, 0 ,0 ,-1 ,12,-2,0,-1,10 ,12,-2,0,-1, 10, 0 ,0 ,-1	,12,-2,0,-1,10 ,12,-2,0,-1, 10, 0 ,0 ,-1 ,12,-2,0,-1,10 ,12,-2,0,-1, 10, 0 ,0 ,-1 ,12,-2,0,-1,10 ,12,-2,0,-1, 10, 0 ,0 ,-1 ,12,-2,0,-1,10 ,12,-2,0,-1, 10, 0 ,0 ,-1 ,12,-2,0,-1,10 ,12,-2,0,-1, 10, 0 ,0 ,-1 ,12,-2,0,-1,10 ,12,-2,0,-1, 10, 0 ,0 ,-1 ,12,-2,0,-1,10 ,12,-2,0,-1, 10, 0 ,0 ,-1 ,12,-2,0,-1,10 ,0,0 ,0,0 ,0,0 ,0,0 ,0,0 ,0,0 ,0,0};

char NSin1[] = { 13,7,7,1,-17,7,7,2,-14,7,7,-2,-16,7,7,-15,7,7,-15,7,			-1,7,7,1,-17,7,7,2,-14,7,7,-2,-16,7,7,-15,7,-8,7,		-13,7,7, 1,-17, 7, 7, 2,-14, 7, 7, -2,-16,7, 7,-15 		,5,7,7,1,-17,7,7,2,-14,7,7,-2,-16,7,7,-15,		5, 7, 7,1,-17, 7,7,2 ,-14,7,7 ,-2,-16,7,7,-15,	5,7, 7,1,-17, 7,7, 2,-14,7, 7,-2,-16,-1,								           17,7,7,1,-17,7,7,2,-14,7,7,-2,-16,7,7,-15,7,7,-15,7,   -1,7,7,1,-17,7,7,2,-14,7,7,-2,-16,7,7,-15,7,7,-15,7,		-1,7,7,1,-17,7,7,2,-14,7,7,-2,-16,7,7,-15,7,7,-15,7,-1,7,7,1,-17,7,7,2,-14,7,7,-2,-16,7,7,-15,7,7,-15,7,0,0,0,0};
char NSqr1[] = {  0,0,0,0,  0,0,0,0,  0,0,0, 0,  0,0,0, 0, 0,0, 0 ,0,			13,0,0,0,-2,0,0, 0,  2,0,0, 0,-4, 0,0,-1 ,0,-1,0,		6,-13,1,-1,0  , 1,-1, 0, 4 ,-4, 1 ,-1 ,0 ,1,-1 ,0,		6, -6, 1 ,-1 ,0,1 ,-1 ,0,4,-4,1,-1,0,1,-1,		6,-5,-1,1 ,-1, 0,1,-1,  0,4,-4, 1,-1 ,0,1,-1 ,  0,4,-4,1,- 1, 0,1,-1, 0, 4,-4, 1, -1,0,1,											12,0,0,0, -2,0,0,0,  2,0,0, 0, -4,0,0,-1 ,0,0,-1,0,		6,0,0,0,-2,0,0, 0,  2,0,0, 0,-4, 0,0,-1 ,0,0,-1,0,		6,0,0,0,-2,0,0, 0,  2,0,0, 0,-4, 0,0,-1 ,0,0,-1,0, 6,-13,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
unsigned char DBase[] = {0,0,0,0,0, 0,0 ,0 ,0 ,0,0,0,0,0 ,0,0 , 0,0,0 ,0    ,0,0 ,0 ,0 ,0, 0, 0, 0,0, 0, 0, 0,0, 0,0,0,0,0,0,					1, 0, 1 ,0,0   ,1, 0, 0, 1,  0,  1,0,0,  1,0,0,			1,  0, 1,0,1,   1,0,0,   1,0, 1,0,0, 1,0,		1, 1,0, 1,0,   0,1,0,0,   1,0,  1,0, 0 ,1, 0,	1,1,1,1,   0, 0,1, 0,  0 ,1,0,  1,0,0,  1,	1, 0 , 0 ,1,0,1,0,0, 1,0, 0 , 0,1,0,0, 1,0,0, 1,1,1, 0 , 0 ,1,0,1,0,0, 1,0,	0 , 0,1,0,0, 1,0,0, 1,1,1, 0 , 0 ,1,0,1,0,0, 1,0, 0 , 0,1,0,0, 1,0,0, 1,1,0,0};
char VSqr1[]=  { 0,0,0,0,0, 0,0 ,0 ,0 ,0,0,0,0,0 ,0,0 ,0 ,0,0 ,0	,1,0, 0 ,0 ,0,0,0,0,0,0,0,0, 0, 1,0,1, 0,0 ,0,0,0,0, 0, 0, 0, 0 ,0 ,0,0,0 ,0 ,0 ,0,0,0,0,0 ,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,-3,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,-1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,-1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,-1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,-1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};


#ifdef USE_SDL
#else
IDirectSound *DSoundPointer = 0;
IDirectSoundBuffer *SoundBuffer = 0;
#endif

void*  WriteHere1;
unsigned long  Length1;
void*  WriteHere2;
unsigned long  Length2;

short fsin1 = 0;
short fsqr = 0;

float volume1 = 0;
float volume2 = 0;
float volume3 = 0;
short RandSeed = 0;
int TimeLeft = 0;
int Position = -1;

long Rendered = -1;
float Notes[36];

void clean_up();

float Sine(float frq, float volume)
{
#ifdef USINGLIBS
	return sinf(frq)*volume;
#else

#ifdef __GNUC__
	asm volatile (
		"fld %1\n\t"
		"fsin\n\t"
		"fmul %2\n\t"
		"fstp %0\n\t"
		: "=m"(volume)
		: "m"(frq), "m"(volume)
	);
#else	// assume MSVC
	__asm{
		fld frq
		fsin
		fmul volume
		fstp volume
	}
	return volume;
#endif	// __GNUC__
#endif
}

float square(float frq, float volume)
{
	float temp = Sine(frq,10);
	if (temp > 0) return volume;
	else return (-volume);
}

float limit(float what)
{
	if (what <0) return 0;
	if (what > 2000) return 2000;
	return what;
}

void Render(short* where,int Length)
{
	int i;
	for (i = 1; i <Length; i+=2){
		if (TimeLeft ==0){
			TimeLeft = NoteLength;
			Position++;
			fsin1 += NSin1[Position];
			fsqr  += NSqr1[Position];
			if (Position >37) TimeLeft -= 11050;
			else
			if ((Position == 3)||(Position == 7)||(Position == 11)||
			   (Position == 23)||(Position == 27)||(Position == 31))
				TimeLeft += NoteLength;
		}

		float frq1 = (int)Notes[fsin1]*(TwoPi*TimeLeft);
		float frq3 = (int)Notes[fsqr]*(TwoPi*TimeLeft);
		float frq4 = (int)Notes[fsqr+7]*(TwoPi*TimeLeft);

		if (frq3==0) frq4 = 0;

		volume1 = limit(volume1 + VSin1[Position]*1000.0/NoteLength);
		volume2 = limit(volume2 + VSin2[Position]*1000.0/NoteLength);
		volume3 = limit(volume3 + VSqr1[Position]*500.0/NoteLength);


		where[i]= 2*square(frq3,volume3)+
					square(frq4,volume3)-
					square(frq3/2,volume3)+
					square(frq3/4,volume3)-
					square(frq4/2,volume3)+
					square(frq3/8,volume3)-
					square(frq4/4,volume3);

		where[i-1] = where[i];

		where[i] += Sine(frq1/2.0,volume1);
		where[i] += Sine(frq1,volume1);
		
		where[i-1] += Sine(frq1/2.0,volume2);
		where[i-1] += Sine(frq1,volume2);

		if(DBase[Position]>0)
		{
			where[i-1] += Sine(97.99*TimeLeft/11050,10*volume3);
			where[i] += Sine(97.99*(TimeLeft*TimeLeft)/(11050*11050),10*volume3);
		}

TimeLeft --;
	}
}

#ifdef USE_SDL
#define MIN(a, b)	((a) < (b) ? (a) : (b))
//static unsigned char *audio_pos;

void audio_callback(void *userdata, unsigned char *stream, int len) {
	Render((short*)stream, len>>1);
}
#endif	// USE_SDL

void CheckPosition()
{
#ifdef USE_SDL
	
#else
	unsigned long Play, Write;
	do {SoundBuffer->GetCurrentPosition(&Play,&Write);
	} while ( Play >= 88200);

	int Length = (int)Play - (int)Rendered;
	if (Length <= 0) Length += 88200;
	if (Length > 44100)
	{
		SoundBuffer->Lock(Rendered,Length,&WriteHere1,&Length1,&WriteHere2,&Length2,NULL);
		Render((short*)WriteHere1,Length1>>1);
		Render((short*)WriteHere2,Length2>>1);
		SoundBuffer->Unlock(WriteHere1,Length1,WriteHere2,Length2);
		Rendered+=Length;
		if (Rendered > 88200) Rendered -= 88200;
	}
#endif	// USE_SDL
}

void mainCRTStartup(void)
{
	char *buf;
	int i;

#ifdef USE_SDL
	SDL_Init(SDL_INIT_VIDEO | SDL_INIT_AUDIO | SDL_INIT_TIMER);
	Backbuffer = SDL_SetVideoMode(WIDTH, HEIGHT, 32, SDL_SWSURFACE | (windowed ? 0 : SDL_FULLSCREEN));

	SDL_WM_SetCaption("Spherez / Pneuma :: GNU/Linux port by Nuclear / The Lab", 0);
#else
	/*
	 * Init DirectDraw
	 */
	DirectDrawCreateEx(NULL, (void**)&DDraw,IID_IDirectDraw7, NULL);
	DDraw->SetCooperativeLevel (CreateWindow("STATIC", "", WS_VISIBLE, 0, 0, WIDTH, HEIGHT, NULL, NULL, NULL, NULL), DDSCL_EXCLUSIVE | DDSCL_FULLSCREEN | DDSCL_ALLOWREBOOT);
	DDraw->SetDisplayMode (WIDTH, HEIGHT, 32, 0, 0);

	//Create Primary buffer with backbuffer
	buf = (char *)&ScreenDesc;
	for (i=sizeof (DDSURFACEDESC2); i; i--)
		*buf++ = 0;
	ScreenDesc.dwSize = sizeof (DDSURFACEDESC2);
	ScreenDesc.dwFlags = DDSD_CAPS | DDSD_BACKBUFFERCOUNT;
	ScreenDesc.dwBackBufferCount = 1;
	ScreenDesc.ddsCaps.dwCaps = DDSCAPS_PRIMARYSURFACE | DDSCAPS_COMPLEX | DDSCAPS_FLIP;
	DDraw->CreateSurface (&ScreenDesc, &Primary, 0);

	//Create BackBuffer
	DDSCAPS2 ddscaps;
	buf = (char *)&ddscaps;
	for (i=sizeof (DDSCAPS2); i; i--)
		*buf++ = 0;
	ddscaps.dwCaps = DDSCAPS_BACKBUFFER;
	Primary->GetAttachedSurface (&ddscaps, &Backbuffer);
#endif	// USE_SDL
	ShowCursor (0);

#ifdef USE_SDL
	TTF_Init();
	if(!(font = TTF_OpenFont("Verdana.ttf", 16))) {
#ifdef USINGLIBS
		fprintf(stderr, "could not open font file, make sure you run this from the binary dir\n");
#endif
		clean_up();
		exit(-1);
	}

#else
	/*
	 * Init fonts
	 */
	BITMAPINFO bmi;
	FontDC = CreateCompatibleDC(NULL);
	SelectObject(FontDC, CreateFont ( FONTHEIGHT-2, 0, 0, 0, FW_NORMAL, FALSE, FALSE, FALSE,
						ANSI_CHARSET, OUT_DEFAULT_PRECIS, CLIP_DEFAULT_PRECIS, ANTIALIASED_QUALITY, VARIABLE_PITCH,
						"Arial"));
	buf = (char *)&bmi;
	for (i=sizeof (BITMAPINFO); i; i--)
		*buf++ = 0;
	bmi.bmiHeader.biSize = sizeof(BITMAPINFOHEADER);
	bmi.bmiHeader.biWidth = WIDTH;
	bmi.bmiHeader.biHeight = -FONTHEIGHT;
	bmi.bmiHeader.biPlanes=1;
	bmi.bmiHeader.biBitCount = 32;
    // Set up our memory DC with the font and bitmap
    SelectObject(FontDC, CreateDIBSection (NULL, &bmi, DIB_RGB_COLORS, (void **) &BmpBits, NULL, 0));
    SetBkColor(FontDC, 0);
    SetTextColor(FontDC, 0x00ffffff);
#endif	// USE_SDL

	Notes[0] = 0;
	Notes[1] = 82.40710208;
	for (i = 2; i < 36 ; i++)
		Notes[i] = Notes[i-1] * 1.059463f;
#ifdef USE_SDL
	SDL_AudioSpec audio;
	audio.freq = 44100;
	audio.format = AUDIO_S16;
	audio.channels = 2;
	audio.samples = 2048;
	audio.callback = audio_callback;
	audio.userdata = 0;

	if(SDL_OpenAudio(&audio, 0) == -1) {
#ifdef USINGLIBS
		fprintf(stderr, "could not open audio device\n");
#endif
		exit(-1);
	}
	SDL_PauseAudio(0);
#else
	//InitSound
	DirectSoundCreate( NULL, &DSoundPointer, NULL);
	DSoundPointer->SetCooperativeLevel(GetForegroundWindow(), DSSCL_NORMAL );
	WAVEFORMATEX wavefmt;
	wavefmt.cbSize = 0;
	wavefmt.wFormatTag = WAVE_FORMAT_PCM;
	wavefmt.nChannels = 2;
	wavefmt.nSamplesPerSec = 44100;
	wavefmt.nBlockAlign = 4; // 16 bit sound data = 2 bytes *2 stereo
	wavefmt.nAvgBytesPerSec = 176400;
	wavefmt.wBitsPerSample = 16;
	DSBUFFERDESC Desc;
	Desc.dwReserved = 0;
	Desc.dwSize = sizeof DSBUFFERDESC;
	Desc.dwFlags = 0;
	Desc.dwBufferBytes = 88200;
	Desc.lpwfxFormat = &wavefmt;
	DSoundPointer->CreateSoundBuffer(&Desc, &(SoundBuffer), NULL);
	SoundBuffer->Play(0,0,DSBPLAY_LOOPING);

	while (Position!=0)
		CheckPosition ();
#endif	// USE_SDL

	/*
	short foo[1024];
	while(Position != 0) {
		Render(foo, 1024);
	}
	*/

	Ray r = { { WIDTH/2, HEIGHT/2, -700 } };
	Vector col;
	int x,y;
	int Start = GetTickCount(), CurTick=0, Offset;
	float tmp1,tmp2;
	float ScrollerY; //apo katw
#ifndef USE_SDL
	SIZE size;
#endif

#ifdef USE_SDL
	int done = 0;
	while(!done && CurTick < 57500) {
		SDL_Event event;
		while(SDL_PollEvent(&event)) {
			if(event.type == SDL_KEYDOWN && event.key.keysym.sym == SDLK_ESCAPE) {
				done = 1;
			}
		}
#else		
	while(GetAsyncKeyState(VK_ESCAPE)==0  && CurTick < 57500 )
	{
#endif	// USE_SDL
		CheckPosition ();
		CurTick = GetTickCount () - Start;


		
		if (CurTick > 18750)
			Offset = 0;
		else if ( CurTick/1250 == 7 )
			goto after;
		else 
			Offset = Off[CurTick/1250];

		tmp1=10.52*(Offset+CurTick)/20000.0f;
after:
#ifdef USINGLIBS
		tmp2 = cosf(tmp1);
		tmp1 = sinf(tmp1);
#else
#ifdef __GNUC__
		asm volatile (
			"fld %2\n\t"
			"fsincos\n\t"
			"fstp %0\n\t"
			"fstp %1\n\t"
			: "=m"(tmp2), "=m"(tmp1)
			: "f"(tmp1)
		);
#else	// assume MSVC
		_asm
		{
			fld tmp1
			fsincos
			fstp tmp2
			fstp tmp1
		}
#endif	// __GNUC__
#endif

		for (i=0; i<5; i++)
		{
			Spheres[i+2].Position.x = 320 + Amp[i].x*(Phase[i][0]?tmp2:tmp1);
			Spheres[i+2].Position.y = 240 + Amp[i].y*(Phase[i][1]?tmp2:tmp1);
			Spheres[i+2].Position.z = 260 + Amp[i].z*(Phase[i][2]?tmp2:tmp1);
		}

		float note = (float)CurTick/250 -50.18;//(CurTick-45*501)/250+40;
		if (note >= 0) if (DBase[(int)note])
			Spheres[2].Radius = Spheres[3].Radius = Spheres[4].Radius = Spheres[5].Radius = Spheres[6].Radius = 72;
		else
			Spheres[2].Radius = Spheres[3].Radius = Spheres[4].Radius = Spheres[5].Radius = Spheres[6].Radius = 36;
		ScrollerY = 0.0705882 * (CurTick-38000);
		
		//Lock
#ifdef USE_SDL
		if(SDL_MUSTLOCK(Backbuffer)) {
			SDL_LockSurface(Backbuffer);
		}
#else
		Backbuffer->Lock (0, &ScreenDesc, DDLOCK_WAIT, 0);
		ScreenDesc.lPitch >>= 2;
#endif

		//Render
		for (y=0; y<HEIGHT; y++)
		{
			for (x=0; x<WIDTH; x++)
			{
				r.b.x = (float)x;
				r.b.y = (float)y;
				r.b.z = 0;

				r.v.x = r.b.x - r.a.x; 
				r.v.y = r.b.y - r.a.y;
				r.v.z = r.b.z - r.a.z;
				
				col.x = col.y = col.z = 0;
				
				Raycast (&r, &col, 2);
				
				if (col.x > 255) col.x = 255;
				if (col.y > 255) col.y = 255;
				if (col.z > 255) col.z = 255;
#ifdef USE_SDL
				buf = (char*)&((unsigned long*)Backbuffer->pixels)[y * (Backbuffer->pitch>>2) + x];
#else
				buf = &((unsigned long *)ScreenDesc.lpSurface)[y*(ScreenDesc.lPitch)+x];
#endif	// USE_SDL
				*buf++ = col.x;
				*buf++ = col.y;
				*buf++ = col.z;
			}
		}
		if (ScrollerY > 0.0f)
		{
			for (i=0; i<LINES; i++)
			{
				int j=0;
				for (;;)
					if (Text[i][j] != '\n')
						j++;
					else
						break;

#ifdef USE_SDL
				SDL_Color color = {255, 255, 255};
				
				char *ptr = Text[i];
				while(*ptr && *ptr != '\n') ptr++;
				if(!*ptr) ptr = 0;
				
				if(ptr) *ptr = 0;
				SDL_Surface *text_surf = TTF_RenderText_Solid(font, Text[i], color);
				if(ptr) *ptr = '\n';

				SDL_Rect rect;
				rect.x = 0;
				rect.y = HEIGHT - (int)ScrollerY + i * FONTHEIGHT;
				SDL_BlitSurface(text_surf, NULL, Backbuffer, &rect);
				SDL_FreeSurface(text_surf);
#else
				PatBlt (FontDC, 0, 0, WIDTH, FONTHEIGHT, BLACKNESS);
				TextOut (FontDC, 0, 0, Text[i], j);

				GetTextExtentPoint32(FontDC, Text[i], j, &size);

				for (y=0; y<FONTHEIGHT; y++) {
					for (x=0; x<WIDTH; x++)
					{
						int j = HEIGHT-int(ScrollerY)+i*FONTHEIGHT+y;
						if (j >= HEIGHT)
							break;
						if (j < 0)
							continue;
						
						if (((unsigned long*)BmpBits)[y*640+x])
							((unsigned long *)ScreenDesc.lpSurface)[j*(ScreenDesc.lPitch)+x] = ((unsigned long*)BmpBits)[y*640+x];

					}
				}
#endif	// USE_SDL
			}
		}

#ifdef USE_SDL
		if(SDL_MUSTLOCK(Backbuffer)) {
			SDL_UnlockSurface(Backbuffer);
		}
		SDL_Flip(Backbuffer);
#else
		//Unlock
		Backbuffer->Unlock (0);

		//Flip
		Primary->Flip (0, DDFLIP_WAIT);
#endif	// USE_SDL
	}

	clean_up();
}

void clean_up() {
#ifdef USE_SDL
	if(font) TTF_CloseFont(font);
	TTF_Quit();
	SDL_CloseAudio();
	SDL_Quit();
#else
	Backbuffer->Release ();
	Primary->Release ();
	DDraw->Release ();
	SoundBuffer->Release ();
	DSoundPointer->Release ();
#endif	// USE_SDL
}
#if defined(USINGLIBS)
int main(int argc, char **argv)
{
	if(argc > 1 && argv[1][0] == '-' && argv[1][1] == 'w' && argv[1][2] == 0) {
		windowed = 1;
	}
	
	mainCRTStartup();
	return 0;
}
#endif
